#!/usr/bin/python3
import re
import os
import numpy
from collections import namedtuple
from matplotlib import pyplot

Result = namedtuple('Result', ['time_mean', 'time_std', 'points_mean', 'points_std'])

def parse_result(filename):
  leaf_sizes = []
  results = {}
  
  with open(filename, 'r') as f:
    for line in f.readlines():
      if '(warmup)' in line:
        continue
      
      found = re.findall(r'leaf_size=(\S+)', line)
      if found:
        leaf_sizes.append(float(found[0]))
        
      if len(leaf_sizes) == 0:
        continue
      
      stat = r'\s*(\S+)\s*\+\-\s*(\S+)\s*'
      found = re.findall(stat, line)
      if len(found) != 2:
        continue
      
      name = line.split(':')[0].strip()
      if name not in results:
        results[name] = []
      
      results[name].append(Result(float(found[0][0]), float(found[0][1]), float(found[1][0]), float(found[1][1])))
  
  return leaf_sizes, results
        

def main():
  results_path = os.path.dirname(__file__) + '/results'
  
  leaf_sizes = None
  raw_results = []
  for filename in os.listdir(results_path):
    found = re.findall(r'downsampling_benchmark_(\d+).txt', filename)    
    if not found:
      continue
    
    leaf_sizes, rets = parse_result(results_path + '/' + filename)
    raw_results.append((int(found[0]), rets))
  
  raw_results = sorted(raw_results, key=lambda x: x[0])

  def summarize(rets):
    time_mean = numpy.array([x.time_mean for x in rets])
    time_std = numpy.array([x.time_std for x in rets])
    points_mean = numpy.array([x.points_mean for x in rets])
    points_std = numpy.array([x.points_std for x in rets])
    return Result(time_mean, time_std, points_mean, points_std)

  results = {}
  results['pcl_voxelgrid'] = summarize(raw_results[0][1]['pcl_voxelgrid'])
  results['pcl_approx_voxelgrid'] = summarize(raw_results[0][1]['pcl_approx_voxelgrid'])
  results['small_voxelgrid'] = summarize(raw_results[0][1]['small_voxelgrid'])
  for num_threads, rets in raw_results:
    results['small_voxelgrid_tbb ({} threads)'.format(num_threads)] = summarize(rets['small_voxelgrid_tbb'])
    results['small_voxelgrid_omp ({} threads)'.format(num_threads)] = summarize(rets['small_voxelgrid_omp'])
  
  fig, axes = pyplot.subplots(1, 5)
  fig.set_size_inches(18, 3)

  leaf_size_indices = [0, 1, 4, 9, -1]
  for i, leaf_size_index in enumerate(leaf_size_indices):
    leaf_size = leaf_sizes[leaf_size_index]

    num_threads = [x[0] for x in raw_results]
    time_tbb = []
    time_omp = []
    for N in [x[0] for x in raw_results]:
      time_tbb.append(results['small_voxelgrid_tbb ({} threads)'.format(N)].time_mean[leaf_size_index])
      time_omp.append(results['small_voxelgrid_omp ({} threads)'.format(N)].time_mean[leaf_size_index])
    
    baseline = results['pcl_voxelgrid'].time_mean[leaf_size_index]
    axes[i].plot([num_threads[0], num_threads[-1]], [baseline, baseline], label='pcl_voxelgrid', linestyle='--')
    axes[i].scatter(num_threads, time_tbb, label='small_voxelgrid_tbb', marker='^')
    axes[i].scatter(num_threads, time_omp, label='small_voxelgrid_omp', marker='^')
    
    axes[i].grid()
    axes[i].set_xscale('log')
    axes[i].set_xlabel('Num threads')
    axes[i].set_title('Leaf size = {} m'.format(leaf_size))
  
  axes[0].set_ylabel('Processing time [msec/scan]')
  axes[2].legend(loc='upper center', bbox_to_anchor=(0.5, -0.11), ncol=3)

  fig.savefig('downsampling_threads.svg')

  
  fig, axes = pyplot.subplots(1, 2)
  fig.set_size_inches(18, 3)

  methods = ['small_voxelgrid', 'small_voxelgrid_tbb (2 threads)', 'small_voxelgrid_tbb (3 threads)', 'small_voxelgrid_tbb (4 threads)', 'small_voxelgrid_tbb (6 threads)', 'pcl_voxelgrid', 'pcl_approx_voxelgrid']
  labels = ['small_voxelgrid', 'small_voxelgrid_tbb (2 threads)', 'small_voxelgrid_tbb (3 threads)', 'small_voxelgrid_tbb (4 threads)', 'small_voxelgrid_tbb (6 threads)', 'pcl_voxelgrid', 'pcl_approx_voxelgrid']
  markers = ['^', '^', '^', '^', '^', 'o', 'o']
  for method, label, marker in zip(methods, labels, markers):
    axes[0].plot(leaf_sizes, results[method].time_mean, label=label, linestyle='--' if 'pcl' in method else '-', marker=marker)
    
    if 'threads' not in method or '2 threads' in method:
      axes[1].plot(leaf_sizes, results[method].points_mean / results['pcl_voxelgrid'].points_mean, label=label, linestyle='--' if 'pcl' in method else '-', marker=marker)
  
  axes[0].set_xlabel('Leaf size [m]')
  axes[0].set_ylabel('Processing time [msec/scan]')
  axes[1].set_xlabel('Leaf size [m]')
  axes[1].set_ylabel('Num points ratio to pcl_voxelgrid')
  axes[0].legend()
  axes[1].legend()
  axes[0].grid()
  axes[1].grid()
  
  fig.savefig('downsampling_comp.svg')
  
  pyplot.show()


if __name__ == '__main__':
  main()



#!/usr/bin/python3
import os
import re
import numpy
from collections import namedtuple
from matplotlib import pyplot

Result = namedtuple('Result', ['reg_mean', 'reg_std', 'tp_mean', 'tp_std'])

def parse_result(filename):
  reg_mean = None
  reg_std = None
  throughput_mean = None
  throughput_std = None
  with open(filename, 'r') as f:
    for line in f.readlines():
      found = re.findall(r'([^=]+)\s*\+\-\s*(\S+)', line)
      if not found or len(found) != 2:
        found = re.findall(r'total_throughput=(\S+)', line)
        if found:
          throughput_mean = float(found[0])
        continue
      
      reg_mean = float(found[0][0].strip())
      reg_std = float(found[0][1].strip())
      throughput_mean = float(found[1][0].strip())
      throughput_std = float(found[1][1].strip())
    
  return Result(reg_mean, reg_std, throughput_mean, throughput_std)

def main():
  results_path = os.path.dirname(__file__) + '/results'

  results = {}
  for filename in os.listdir(results_path):
    found = re.findall(r'odometry_benchmark_(\S+)_(native|nonnative)_(\d+).txt', filename)
    if not found:
      continue    
   
    rets = parse_result(results_path + '/' + filename)
    results['{}_{}_{}'.format(found[0][0], found[0][1], found[0][2])] = rets
  
  fig, axes = pyplot.subplots(1, 1, figsize=(12, 2))
  axes = [axes]

  num_threads = [1, 2, 4, 8, 16, 32, 64, 128]

  print(results['small_gicp_native_1'], results['small_gicp_tbb_native_1'])
  print(results['small_gicp_nonnative_1'], results['small_gicp_tbb_nonnative_1'])
  
  native = [results['small_gicp_tbb_native_{}'.format(N)].reg_mean for N in num_threads]
  nonnative = [results['small_gicp_tbb_nonnative_{}'.format(N)].reg_mean for N in num_threads]

  axes[0].plot(num_threads, native, label='small_gicp_tbb (-march=native)', marker='o')
  axes[0].plot(num_threads, nonnative, label='small_gicp_tbb (nonnative)', marker='o')
  axes[0].set_xlabel('Number of threads [1, 2, ..., 128]')
  axes[0].set_ylabel('Time [msec/scan]')
  axes[0].set_xscale('log')
  axes[0].grid()
  axes[0].legend()
  
  pyplot.show()    
    

if __name__ == "__main__":
  main()


#!/usr/bin/python3
# SPDX-FileCopyrightText: Copyright 2024 Kenji Koide
# SPDX-License-Identifier: MIT
import numpy
from scipy.spatial.transform import Rotation

import small_gicp


# Basic registation example with numpy arrays
def example_numpy1(target_raw_numpy : numpy.ndarray, source_raw_numpy : numpy.ndarray):
  # Example A : Perform registration with numpy arrays
  # Arguments
  # - target_points               : Nx4 or Nx3 numpy array of the target point cloud
  # - source_points               : Nx4 or Nx3 numpy array of the source point cloud
  # Optional arguments
  # - init_T_target_source        : Initial guess of the transformation matrix (4x4 numpy array)
  # - registration_type           : Registration type ("ICP", "PLANE_ICP", "GICP", "VGICP")
  # - voxel_resolution            : Voxel resolution for VGICP
  # - downsampling_resolution     : Downsampling resolution
  # - max_correspondence_distance : Maximum correspondence distance
  # - num_threads                 : Number of threads
  result = small_gicp.align(target_raw_numpy, source_raw_numpy, downsampling_resolution=0.25)

  return result.T_target_source

# Example to perform preprocessing and registration separately
def example_numpy2(target_raw_numpy : numpy.ndarray, source_raw_numpy : numpy.ndarray):
  # Example B : Perform preprocessing and registration separately

  # Preprocess point clouds
  # Arguments
  # - points                      : Nx4 or Nx3 numpy array of the target point cloud
  # Optional arguments
  # - downsampling_resolution     : Downsampling resolution
  # - num_neighbors               : Number of neighbors for normal and covariance estimation
  # - num_threads                 : Number of threads
  target, target_tree = small_gicp.preprocess_points(target_raw_numpy, downsampling_resolution=0.25)
  source, source_tree = small_gicp.preprocess_points(source_raw_numpy, downsampling_resolution=0.25)

  # Align point clouds
  # Arguments
  # - target                      : Target point cloud (small_gicp.PointCloud)
  # - source                      : Source point cloud (small_gicp.PointCloud)
  # - target_tree                 : KD-tree of the target point cloud
  # Optional arguments
  # - init_T_target_source        : Initial guess of the transformation matrix (4x4 numpy array)
  # - max_correspondence_distance : Maximum correspondence distance
  # - num_threads                 : Number of threads
  result = small_gicp.align(target, source, target_tree)
  
  return result.T_target_source


# Basic registation example with small_gicp.PointCloud
def example_small1(target_raw_numpy : numpy.ndarray, source_raw_numpy : numpy.ndarray):
  # Convert numpy arrays (Nx3 or Nx4) to small_gicp.PointCloud
  target_raw = small_gicp.PointCloud(target_raw_numpy)
  source_raw = small_gicp.PointCloud(source_raw_numpy)

  # Preprocess point clouds
  target, target_tree = small_gicp.preprocess_points(target_raw, downsampling_resolution=0.25)
  source, source_tree = small_gicp.preprocess_points(source_raw, downsampling_resolution=0.25)
  
  result = small_gicp.align(target, source, target_tree)
  
  return result.T_target_source
  
# Example to perform each preprocessing and registration separately
def example_small2(target_raw_numpy : numpy.ndarray, source_raw_numpy : numpy.ndarray):
  # Convert numpy arrays (Nx3 or Nx4) to small_gicp.PointCloud
  target_raw = small_gicp.PointCloud(target_raw_numpy)
  source_raw = small_gicp.PointCloud(source_raw_numpy)

  # Downsampling
  target = small_gicp.voxelgrid_sampling(target_raw, 0.25)
  source = small_gicp.voxelgrid_sampling(source_raw, 0.25)
  
  # KdTree construction
  target_tree = small_gicp.KdTree(target)
  source_tree = small_gicp.KdTree(source)
  
  # Estimate covariances
  small_gicp.estimate_covariances(target, target_tree)
  small_gicp.estimate_covariances(source, source_tree)

  # Align point clouds  
  result = small_gicp.align(target, source, target_tree)
  
  return result.T_target_source


### Following functions are for testing ###

# Verity the estimated transformation matrix (for testing)
def verify_result(T_target_source, gt_T_target_source):
  error = numpy.linalg.inv(T_target_source) @ gt_T_target_source
  error_trans = numpy.linalg.norm(error[:3, 3])
  error_rot = Rotation.from_matrix(error[:3, :3]).magnitude()
  
  assert error_trans < 0.05
  assert error_rot < 0.05

import pytest

# Load the point clouds and the ground truth transformation matrix
@pytest.fixture(scope='module', autouse=True)
def load_points():
  gt_T_target_source = numpy.loadtxt('data/T_target_source.txt')  # Load the ground truth transformation matrix
  print('--- gt_T_target_source ---')
  print(gt_T_target_source)

  target_raw = small_gicp.read_ply(('data/target.ply'))  # Read the target point cloud (small_gicp.PointCloud)
  source_raw = small_gicp.read_ply(('data/source.ply'))  # Read the source point cloud (small_gicp.PointCloud)

  target_raw_numpy = target_raw.points()                    # Nx4 numpy array of the target point cloud
  source_raw_numpy = source_raw.points()                    # Nx4 numpy array of the source point cloud
  
  yield (gt_T_target_source, target_raw_numpy, source_raw_numpy)

# Check if the point clouds are loaded correctly
def test_load_points(load_points):
  gt_T_target_source, target_raw_numpy, source_raw_numpy = load_points
  assert gt_T_target_source.shape[0] == 4 and gt_T_target_source.shape[1] == 4
  assert len(target_raw_numpy) > 0
  assert len(source_raw_numpy) > 0

def test_example_numpy1(load_points):
  gt_T_target_source, target_raw_numpy, source_raw_numpy = load_points
  T_target_source = example_numpy1(target_raw_numpy, source_raw_numpy)
  verify_result(T_target_source, gt_T_target_source)

def test_example_numpy2(load_points):
  gt_T_target_source, target_raw_numpy, source_raw_numpy = load_points
  T_target_source = example_numpy2(target_raw_numpy, source_raw_numpy)
  verify_result(T_target_source, gt_T_target_source)

def test_example_small1(load_points):
  gt_T_target_source, target_raw_numpy, source_raw_numpy = load_points
  T_target_source = example_small1(target_raw_numpy, source_raw_numpy)
  verify_result(T_target_source, gt_T_target_source)

def test_example_small2(load_points):
  gt_T_target_source, target_raw_numpy, source_raw_numpy = load_points
  T_target_source = example_small2(target_raw_numpy, source_raw_numpy)
  verify_result(T_target_source, gt_T_target_source)

if __name__ == "__main__":
  target_raw = small_gicp.read_ply(('data/target.ply'))  # Read the target point cloud (small_gicp.PointCloud)
  source_raw = small_gicp.read_ply(('data/source.ply'))  # Read the source point cloud (small_gicp.PointCloud)

  target_raw_numpy = target_raw.points()                    # Nx4 numpy array of the target point cloud
  source_raw_numpy = source_raw.points()                    # Nx4 numpy array of the source point cloud

  T_target_source = example_numpy1(target_raw_numpy, source_raw_numpy)
  T_target_source = example_numpy2(target_raw_numpy, source_raw_numpy)
  T_target_source = example_small1(target_raw_numpy, source_raw_numpy)
  T_target_source = example_small2(target_raw_numpy, source_raw_numpy)



#!/usr/bin/python3
# SPDX-FileCopyrightText: Copyright 2024 Kenji Koide
# SPDX-License-Identifier: MIT
import numpy
from scipy.spatial.transform import Rotation

import small_gicp


# Basic registation example with small_gicp.PointCloud
def example_small1(target_raw_numpy : numpy.ndarray, source_raw_numpy : numpy.ndarray):
  # Convert numpy arrays (Nx3 or Nx4) to small_gicp.PointCloud
  target_raw = small_gicp.PointCloud(target_raw_numpy)
  source_raw = small_gicp.PointCloud(source_raw_numpy)

  # Preprocess point clouds
  target, target_tree = small_gicp.preprocess_points(target_raw, downsampling_resolution=0.25)
  source, source_tree = small_gicp.preprocess_points(source_raw, downsampling_resolution=0.25)
  
  result = small_gicp.align(target, source, target_tree)
  
  return result.T_target_source
  
# Example to perform each preprocessing and registration separately
def example_small2(target_raw_numpy : numpy.ndarray, source_raw_numpy : numpy.ndarray):
  # Convert numpy arrays (Nx3 or Nx4) to small_gicp.PointCloud
  target_raw = small_gicp.PointCloud(target_raw_numpy)
  source_raw = small_gicp.PointCloud(source_raw_numpy)

  # Downsampling
  target = small_gicp.voxelgrid_sampling(target_raw, 0.25)
  source = small_gicp.voxelgrid_sampling(source_raw, 0.25)
  
  # KdTree construction
  target_tree = small_gicp.KdTree(target)
  source_tree = small_gicp.KdTree(source)
  
  # Estimate covariances
  small_gicp.estimate_covariances(target, target_tree)
  small_gicp.estimate_covariances(source, source_tree)

  # Align point clouds  
  result = small_gicp.align(target, source, target_tree)
  
  return result.T_target_source


### Following functions are for testing ###

# Verity the estimated transformation matrix (for testing)
def verify_result(T_target_source, gt_T_target_source):
  error = numpy.linalg.inv(T_target_source) @ gt_T_target_source
  error_trans = numpy.linalg.norm(error[:3, 3])
  error_rot = Rotation.from_matrix(error[:3, :3]).magnitude()
  
  assert error_trans < 0.05
  assert error_rot < 0.05

import pytest

# Load the point clouds and the ground truth transformation matrix
@pytest.fixture(scope='module', autouse=True)
def load_points():
  gt_T_target_source = numpy.loadtxt('data/T_target_source.txt')  # Load the ground truth transformation matrix
  target_raw = small_gicp.read_ply(('data/target.ply'))  # Read the target point cloud (small_gicp.PointCloud)
  source_raw = small_gicp.read_ply(('data/source.ply'))  # Read the source point cloud (small_gicp.PointCloud)

  target_raw_numpy = target_raw.points()                    # Nx4 numpy array of the target point cloud
  source_raw_numpy = source_raw.points()                    # Nx4 numpy array of the source point cloud
  
  yield (gt_T_target_source, target_raw_numpy, source_raw_numpy)

# Check if the point clouds are loaded correctly
def test_load_points(load_points):
  gt_T_target_source, target_raw_numpy, source_raw_numpy = load_points
  assert gt_T_target_source.shape[0] == 4 and gt_T_target_source.shape[1] == 4
  assert len(target_raw_numpy) > 0 and target_raw_numpy.shape[1] == 4
  assert len(source_raw_numpy) > 0 and source_raw_numpy.shape[1] == 4

# Basic point cloud test
def test_points(load_points):
  _, points_numpy, _ = load_points

  points = small_gicp.PointCloud(points_numpy)
  assert points.size() == points_numpy.shape[0]
  assert numpy.all(numpy.abs(points.points() - points_numpy) < 1e-6)

  points = small_gicp.PointCloud(points_numpy[:, :3])
  assert points.size() == points_numpy.shape[0]
  assert numpy.all(numpy.abs(points.points() - points_numpy) < 1e-6)
  
  for i in range(10):
    assert numpy.all(numpy.abs(points.point(i) - points_numpy[i]) < 1e-6)
  

# Downsampling test
def test_downsampling(load_points):
  _, points_numpy, _ = load_points

  downsampled = small_gicp.voxelgrid_sampling(points_numpy, 0.25)
  assert downsampled.size() > 0
    
  downsampled2 = small_gicp.voxelgrid_sampling(points_numpy, 0.25, num_threads=2)
  assert abs(1.0 - downsampled.size() / downsampled2.size()) < 0.05
  
  downsampled2 = small_gicp.voxelgrid_sampling(small_gicp.PointCloud(points_numpy), 0.25)
  assert downsampled.size() == downsampled2.size()
  
  downsampled2 = small_gicp.voxelgrid_sampling(small_gicp.PointCloud(points_numpy), 0.25, num_threads=2)
  assert abs(1.0 - downsampled.size() / downsampled2.size()) < 0.05

# Preprocess test
def test_preprocess(load_points):
  _, points_numpy, _ = load_points

  downsampled, _ = small_gicp.preprocess_points(points_numpy, downsampling_resolution=0.25)
  assert downsampled.size() > 0

  downsampled2, _ = small_gicp.preprocess_points(points_numpy, downsampling_resolution=0.25, num_threads=2)
  assert abs(1.0 - downsampled.size() / downsampled2.size()) < 0.05
  
  downsampled2, _ = small_gicp.preprocess_points(small_gicp.PointCloud(points_numpy), downsampling_resolution=0.25)
  assert downsampled.size() == downsampled2.size()
  
  downsampled2, _ = small_gicp.preprocess_points(small_gicp.PointCloud(points_numpy), downsampling_resolution=0.25, num_threads=2)
  assert abs(1.0 - downsampled.size() / downsampled2.size()) < 0.05

# Voxelmap test
def test_voxelmap(load_points):
  _, points_numpy, _ = load_points

  downsampled = small_gicp.voxelgrid_sampling(points_numpy, 0.25)
  small_gicp.estimate_covariances(downsampled)

  voxelmap = small_gicp.GaussianVoxelMap(0.5)
  voxelmap.insert(downsampled)
  
  assert voxelmap.size() > 0
  assert voxelmap.size() == len(voxelmap)

# Factor test
def test_factors(load_points):
  gt_T_target_source, target_raw_numpy, source_raw_numpy = load_points

  target, target_tree = small_gicp.preprocess_points(target_raw_numpy, downsampling_resolution=0.25)
  source, source_tree = small_gicp.preprocess_points(source_raw_numpy, downsampling_resolution=0.25)

  result = small_gicp.align(target, source, target_tree, gt_T_target_source)
  result = small_gicp.align(target, source, target_tree, result.T_target_source)

  factors = [small_gicp.GICPFactor()]
  rejector = small_gicp.DistanceRejector()

  sum_H = numpy.zeros((6, 6))
  sum_b = numpy.zeros(6)
  sum_e = 0.0

  for i in range(source.size()):
    succ, H, b, e = factors[0].linearize(target, source, target_tree, result.T_target_source, i, rejector)
    if succ:
      sum_H += H
      sum_b += b
      sum_e += e

  assert numpy.max(numpy.abs(result.H - sum_H) / result.H) < 0.05

# Registration test
def test_registration(load_points):
  gt_T_target_source, target_raw_numpy, source_raw_numpy = load_points

  result = small_gicp.align(target_raw_numpy, source_raw_numpy, downsampling_resolution=0.25)
  verify_result(result.T_target_source, gt_T_target_source)

  result = small_gicp.align(target_raw_numpy, source_raw_numpy, downsampling_resolution=0.25, num_threads=2)
  verify_result(result.T_target_source, gt_T_target_source)

  target, target_tree = small_gicp.preprocess_points(target_raw_numpy, downsampling_resolution=0.25)
  source, source_tree = small_gicp.preprocess_points(source_raw_numpy, downsampling_resolution=0.25)

  result = small_gicp.align(target, source)
  verify_result(result.T_target_source, gt_T_target_source)

  result = small_gicp.align(target, source, target_tree)
  verify_result(result.T_target_source, gt_T_target_source)

  target_voxelmap = small_gicp.GaussianVoxelMap(0.5)
  target_voxelmap.insert(target)
  
  result = small_gicp.align(target_voxelmap, source)
  verify_result(result.T_target_source, gt_T_target_source)



#!/usr/bin/python3
import os
import re
import numpy
from collections import namedtuple
from matplotlib import pyplot

Result = namedtuple('Result', ['reg_mean', 'reg_std', 'tp_mean', 'tp_std'])

def parse_result(filename):
  reg_mean = None
  reg_std = None
  throughput_mean = None
  throughput_std = None
  with open(filename, 'r') as f:
    for line in f.readlines():
      found = re.findall(r'([^=]+)\s*\+\-\s*(\S+)', line)
      if not found or len(found) != 2:
        found = re.findall(r'total_throughput=(\S+)', line)
        if found:
          throughput_mean = float(found[0])
        continue
      
      reg_mean = float(found[0][0].strip())
      reg_std = float(found[0][1].strip())
      throughput_mean = float(found[1][0].strip())
      throughput_std = float(found[1][1].strip())
    
  return Result(reg_mean, reg_std, throughput_mean, throughput_std)

def main():
  results_path = os.path.dirname(__file__) + '/results'

  results = {}
  for filename in os.listdir(results_path):
    found = re.findall(r'odometry_benchmark_(\S+)_(\d+).txt', filename)
    if not found:
      continue    
   
    rets = parse_result(results_path + '/' + filename)
    results['{}_{}'.format(found[0][0], found[0][1])] = rets
  
  fig, axes = pyplot.subplots(2, 2, figsize=(24, 12))
  
  num_threads = [1, 2, 4, 8, 16, 32, 64, 128]

  pcl_reg = results['pcl_1'].reg_mean
  pcl_tp = results['pcl_1'].tp_mean
  axes[0, 0].plot([num_threads[0], num_threads[-1]], [pcl_reg, pcl_reg], label='pcl_gicp', linestyle='--')
  axes[0, 1].plot([num_threads[0], num_threads[-1]], [pcl_tp, pcl_tp], label='pcl_gicp', linestyle='--')
  axes[1, 0].plot([num_threads[0], num_threads[-1]], [1.0, 1.0], label='pcl_gicp', linestyle='--')
  axes[1, 1].plot([num_threads[0], num_threads[-1]], [1.0, 1.0], label='pcl_gicp', linestyle='--')

  methods = ['fast_gicp', 'fast_vgicp', 'small_gicp_omp', 'small_gicp_tbb', 'small_vgicp_tbb', 'small_vgicp_omp']
  markers = ['o', 'o', '^', '^', 's', 's']

  for method, marker in zip(methods, markers):
    reg_means = [results['{}_{}'.format(method, N)].reg_mean for N in num_threads]
    axes[0, 0].plot(num_threads, reg_means, label=method, marker=marker)
    axes[1, 0].plot(num_threads, pcl_reg / numpy.array(reg_means), label=method, marker=marker)

  for method, marker in zip(methods, markers):
    tp_means = [results['{}_{}'.format(method, N)].tp_mean for N in num_threads]
    axes[0, 1].plot(num_threads, tp_means, label=method, marker=marker)
    axes[1, 1].plot(num_threads, pcl_tp / numpy.array(tp_means), label=method, marker=marker)
  flow_tp_means = [results['small_gicp_tbb_flow_{}'.format(N)].tp_mean for N in num_threads]
  axes[0, 1].plot(num_threads, flow_tp_means, label='small_gicp_tbb_flow', marker='*')
  axes[1, 1].plot(num_threads, pcl_tp / numpy.array(flow_tp_means), label='small_gicp_tbb_flow', marker='*')

  axes[0, 0].set_title('Net registration time (KdTree construction + cov estimation + pose estimation)')
  axes[1, 0].set_title('Net registration time (KdTree construction + cov estimation + pose estimation)')
  axes[0, 1].set_title('Total throughput (Downsampling + registration)')
  axes[1, 1].set_title('Total throughput (Downsampling + registration)')
  axes[0, 0].set_ylabel('Time [msec/scan]')
  axes[0, 1].set_ylabel('Time [msec/scan]')
  axes[1, 0].set_ylabel('Processing speed ratio (pcl_gicp=1.0)')
  axes[1, 1].set_ylabel('Processing speed ratio (pcl_gicp=1.0)')
  for i in range(2):
    for j in range(2):
      axes[i, j].set_xlabel('Number of threads = [1, 2, 4, ..., 128]')
      axes[i, j].set_xscale('log')
      axes[i, j].legend()
      axes[i, j].grid()
  
  fig.savefig('odometry_time.svg')
  pyplot.show()

if __name__ == "__main__":
  main()